--- ngx_http_xslt_filter_module.c-ori	2012-04-17 15:41:28.000000000 +0200
+++ ngx_http_xslt_filter_module.c	2012-04-18 15:46:26.000000000 +0200
@@ -47,13 +47,20 @@
 
 
 typedef struct {
+    ngx_str_t                  name;
+    xmlDtdPtr                  dtd;
+} ngx_http_xslt_dtd_t;
+
+
+typedef struct {
+    ngx_str_t                  name;
     xsltStylesheetPtr          stylesheet;
     ngx_array_t                params;       /* ngx_http_xslt_param_t */
 } ngx_http_xslt_sheet_t;
 
 
 typedef struct {
-    xmlDtdPtr                  dtd;
+    ngx_http_xslt_dtd_t       *entities;
     ngx_array_t                sheets;       /* ngx_http_xslt_sheet_t */
     ngx_hash_t                 types;
     ngx_array_t               *types_keys;
@@ -415,13 +421,18 @@
 
     doc = ctxt->myDoc;
 
+    if (!conf->entities) {
+        doc->extSubset = NULL;
+        return;
+    }
+
 #if (NGX_HTTP_XSLT_REUSE_DTD)
 
-    dtd = conf->dtd;
+    dtd = conf->entities->dtd;
 
 #else
 
-    dtd = xmlCopyDtd(conf->dtd);
+    dtd = xmlCopyDtd(conf->entities->dtd);
     if (dtd == NULL) {
         ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,
                       "xmlCopyDtd() failed");
@@ -795,49 +806,19 @@
     ngx_http_xslt_filter_loc_conf_t *xlcf = conf;
 
     ngx_str_t                         *value;
-    ngx_uint_t                         i;
-    ngx_pool_cleanup_t                *cln;
-    ngx_http_xslt_file_t              *file;
-    ngx_http_xslt_filter_main_conf_t  *xmcf;
 
-    if (xlcf->dtd) {
+    if (xlcf->entities) {
         return "is duplicate";
     }
 
     value = cf->args->elts;
 
-    xmcf = ngx_http_conf_get_module_main_conf(cf, ngx_http_xslt_filter_module);
-
-    file = xmcf->dtd_files.elts;
-    for (i = 0; i < xmcf->dtd_files.nelts; i++) {
-        if (ngx_strcmp(file[i].name, &value[1].data) == 0) {
-            xlcf->dtd = file[i].data;
-            return NGX_CONF_OK;
-        }
-    }
-
-    cln = ngx_pool_cleanup_add(cf->pool, 0);
-    if (cln == NULL) {
-        return NGX_CONF_ERROR;
-    }
-
-    xlcf->dtd = xmlParseDTD(NULL, (xmlChar *) value[1].data);
-
-    if (xlcf->dtd == NULL) {
-        ngx_conf_log_error(NGX_LOG_ERR, cf, 0, "xmlParseDTD() failed");
-        return NGX_CONF_ERROR;
-    }
-
-    cln->handler = ngx_http_xslt_cleanup_dtd;
-    cln->data = xlcf->dtd;
-
-    file = ngx_array_push(&xmcf->dtd_files);
-    if (file == NULL) {
+    xlcf->entities = ngx_pcalloc(cf->pool, sizeof(ngx_http_xslt_dtd_t));
+    if (xlcf->entities == NULL) {
         return NGX_CONF_ERROR;
     }
 
-    file->name = value[1].data;
-    file->data = xlcf->dtd;
+    xlcf->entities->name = value[1];
 
     return NGX_CONF_OK;
 }
@@ -851,12 +832,9 @@
 
     ngx_str_t                         *value;
     ngx_uint_t                         i, n;
-    ngx_pool_cleanup_t                *cln;
-    ngx_http_xslt_file_t              *file;
     ngx_http_xslt_sheet_t             *sheet;
     ngx_http_xslt_param_t             *param;
     ngx_http_compile_complex_value_t   ccv;
-    ngx_http_xslt_filter_main_conf_t  *xmcf;
 
     value = cf->args->elts;
 
@@ -876,45 +854,7 @@
 
     ngx_memzero(sheet, sizeof(ngx_http_xslt_sheet_t));
 
-    if (ngx_conf_full_name(cf->cycle, &value[1], 0) != NGX_OK) {
-        return NGX_CONF_ERROR;
-    }
-
-    xmcf = ngx_http_conf_get_module_main_conf(cf, ngx_http_xslt_filter_module);
-
-    file = xmcf->sheet_files.elts;
-    for (i = 0; i < xmcf->sheet_files.nelts; i++) {
-        if (ngx_strcmp(file[i].name, &value[1].data) == 0) {
-            sheet->stylesheet = file[i].data;
-            goto found;
-        }
-    }
-
-    cln = ngx_pool_cleanup_add(cf->pool, 0);
-    if (cln == NULL) {
-        return NGX_CONF_ERROR;
-    }
-
-    sheet->stylesheet = xsltParseStylesheetFile(value[1].data);
-    if (sheet->stylesheet == NULL) {
-        ngx_conf_log_error(NGX_LOG_ERR, cf, 0,
-                           "xsltParseStylesheetFile(\"%s\") failed",
-                           value[1].data);
-        return NGX_CONF_ERROR;
-    }
-
-    cln->handler = ngx_http_xslt_cleanup_stylesheet;
-    cln->data = sheet->stylesheet;
-
-    file = ngx_array_push(&xmcf->sheet_files);
-    if (file == NULL) {
-        return NGX_CONF_ERROR;
-    }
-
-    file->name = value[1].data;
-    file->data = sheet->stylesheet;
-
-found:
+    sheet->name = value[1];
 
     n = cf->args->nelts;
 
@@ -1067,12 +1007,102 @@
     ngx_http_xslt_filter_loc_conf_t *prev = parent;
     ngx_http_xslt_filter_loc_conf_t *conf = child;
 
-    if (conf->dtd == NULL) {
-        conf->dtd = prev->dtd;
+    ngx_uint_t                         i, n;
+    ngx_pool_cleanup_t                *cln;
+    ngx_http_xslt_file_t              *file;
+    ngx_http_xslt_sheet_t             *sheet;
+    ngx_http_xslt_filter_main_conf_t  *xmcf;
+
+    xmcf = ngx_http_conf_get_module_main_conf(cf, ngx_http_xslt_filter_module);
+
+    if (conf->entities == NULL) {
+        conf->entities = prev->entities;
+
+    }
+    else {
+        if (ngx_conf_full_name(cf->cycle, &conf->entities->name, 0) != NGX_OK) {
+            return NGX_CONF_ERROR;
+        }
+
+        file = xmcf->dtd_files.elts;
+        for (i = 0; i < xmcf->dtd_files.nelts; i++) {
+            if (ngx_strcmp(file[i].name, conf->entities->name.data) == 0) {
+                conf->entities->dtd = file[i].data;
+                return NGX_CONF_OK;
+            }
+        }
+
+        cln = ngx_pool_cleanup_add(cf->pool, 0);
+        if (cln == NULL) {
+            return NGX_CONF_ERROR;
+        }
+
+        conf->entities->dtd = xmlParseDTD(NULL,
+                                         (xmlChar *) conf->entities->name.data);
+
+        if (conf->entities->dtd == NULL) {
+            ngx_conf_log_error(NGX_LOG_ERR, cf, 0, "xmlParseDTD(\"%s\") failed",
+                               conf->entities->name.data);
+            return NGX_CONF_ERROR;
+        }
+
+        cln->handler = ngx_http_xslt_cleanup_dtd;
+        cln->data = conf->entities->dtd;
+
+        file = ngx_array_push(&xmcf->dtd_files);
+        if (file == NULL) {
+            return NGX_CONF_ERROR;
+        }
+
+        file->name = conf->entities->name.data;
+        file->data = conf->entities->dtd;
     }
 
     if (conf->sheets.nelts == 0) {
         conf->sheets = prev->sheets;
+
+    } else {
+        sheet = conf->sheets.elts;
+        for (n = 0; n < conf->sheets.nelts; n++) {
+            if (ngx_conf_full_name(cf->cycle, &sheet[n].name, 0) != NGX_OK) {
+                return NGX_CONF_ERROR;
+            }
+
+            file = xmcf->sheet_files.elts;
+            for (i = 0; i < xmcf->sheet_files.nelts; i++) {
+                if (ngx_strcmp(file[i].name, sheet[n].name.data) == 0) {
+                    sheet[n].stylesheet = file[i].data;
+                    goto next;
+                }
+            }
+
+            cln = ngx_pool_cleanup_add(cf->pool, 0);
+            if (cln == NULL) {
+                return NGX_CONF_ERROR;
+            }
+
+            sheet[n].stylesheet = xsltParseStylesheetFile(sheet[n].name.data);
+            if (sheet[n].stylesheet == NULL) {
+                ngx_conf_log_error(NGX_LOG_ERR, cf, 0,
+                                   "xsltParseStylesheetFile(\"%s\") failed",
+                                   sheet[n].name.data);
+                return NGX_CONF_ERROR;
+            }
+
+            cln->handler = ngx_http_xslt_cleanup_stylesheet;
+            cln->data = sheet[n].stylesheet;
+
+            file = ngx_array_push(&xmcf->sheet_files);
+            if (file == NULL) {
+                return NGX_CONF_ERROR;
+            }
+
+            file->name = sheet[n].name.data;
+            file->data = sheet[n].stylesheet;
+next:
+
+            continue;
+        }
     }
 
     if (conf->params == NULL) {